@@ -1,5 +1,5 @@
 ;; scss.scm: main module exports and implementations for SCSS
-;; Copyright (C) 2006 Julian Graham
+;; Copyright (C) 2007 Julian Graham
 
 ;; SCSS is free software; you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -54,10 +54,15 @@
 
 (use-modules (ice-9 regex) (srfi srfi-1))
 
-(define unicode-regex "(\\\\[0-9a-f]{1,6}(\\r\\n|[ \\n\\r\\t\\f])?)")
+;; The format of these regexes needs to stay consistent with the code for
+;; replace-escapes
+
+(define unicode-regex "(\\\\([0-9a-fA-F]{1,6})[[:space:]]?)")
 (define unicode-regexc (make-regexp unicode-regex))
-(define escape-regex (string-append "(" unicode-regex "|(\\\\[^nrf]))"))
+(define escape-regex 
+  (string-append "(" unicode-regex "|(\\\\[^[:space:]0-9a-fA-F]))"))
 (define escape-regexc (make-regexp escape-regex))
+
 (define nonascii-regex "([^[:alnum:][:punct:]])")
 (define nonascii-regexc (make-regexp nonascii-regex))
 (define nmstart-regex 
@@ -228,9 +233,10 @@
     ("background-repeat" ,(matchf prepeat-regex) ("repeat") #f)
     ("background"
      ,(lambda (x)
-	(or (any-order-or-match pcolor-regex pimage-regex prepeat-regex
-				pattach-regex lposition)
-	    (matchf "inherit")))
+	(or (apply (any-order-or-match pcolor-regex pimage-regex prepeat-regex
+				       pattach-regex lposition)
+		   (list x))
+	    (exact-match "inherit" x)))
      ("black" "none" "no-repeat" "fixed" "0% 0%") #f)
     ("border-collapse" ,(matchf "(collapse)|(separate)|(inherit)")
      ("separate") #t)
@@ -246,9 +252,10 @@
      ("none" "none" "none" "none") #f)
     (,(list "border-top" "border-right" "border-bottom" "border-left")
      ,(lambda (x) 
-	(or (any-order-or-match border-width-regex 
-				border-style-regex 
-				pcolor-regex)
+	(or (apply (any-order-or-match border-width-regex 
+				       border-style-regex 
+				       pcolor-regex)
+		   (list x))
 	    (exact-match "inherit" x)))
      ("medium" "none" color) #f)
     (,(list "border-top-color" "border-right-color"
@@ -265,9 +272,10 @@
      ("medium" "medium" "medium" "medium") #f)
     ("border"
      ,(lambda (x) 
-	(or (any-order-or-match border-width-regex 
-				border-style-regex 
-				pcolor-regex)
+	(or (apply (any-order-or-match border-width-regex 
+				       border-style-regex 
+				       pcolor-regex)
+		   (list x))
 	    (exact-match "inherit" x)))
      ("medium none" color) #f)
     ("bottom" ,(matchf (string-append length-regex "|" percentage-regex 
@@ -328,8 +336,9 @@
      ,(lambda (x)
 	(let ((y (string-append fsize-regex "(\\w+\\/" lheight-regex ")?\\w+"
 				ffamily-regex)))
-	  (or (and (any-order-or-match fstyle-regex fvariant-regex 
-				       fweight-regex y)
+	  (or (and (apply (any-order-or-match fstyle-regex fvariant-regex 
+					      fweight-regex y)
+			  (list x))
 		   (string-match (string-append y "$") x))
 	      (exact-match y x)
 	      (exact-match (string-append "(caption)|(icon)|(menu)|"
@@ -355,7 +364,9 @@
      ,(matchf (string-append lstype-regex "|(inherit)")) ("disc") #t)
     ("list-style"
      ,(lambda (x)
-	(or (any-order-or-match lstype-regex lsposition-regex lsimage-regex)
+	(or (apply (any-order-or-match lstype-regex lsposition-regex 
+				       lsimage-regex)
+		   (list x))
 	    (exact-match "inherit" x)))
      ("disc" "outside" "none") #t)
     (,(list "margin-right" "margin-left" "margin-top" "margin-bottom")
@@ -380,7 +391,8 @@
      ("medium") #f)
     ("outline" 
      ,(lambda (x) 
-	(or (any-order-or-match ocolor-regex ostyle-regex owidth-regex)
+	(or (apply (any-order-or-match ocolor-regex ostyle-regex owidth-regex)
+		   (list x))
 	    (exact-match "inherit" x)))
      ("invert" "none" "medium") #f)
     ("overflow"
@@ -435,8 +447,9 @@
     ("text-decoration"
      ,(lambda (x) 
 	(or (exact-match "none" x)
-	    ((any-order-or-match "underline" "overline" "line-through" "blink")
-	     x)
+	    (apply (any-order-or-match "underline" "overline" "line-through" 
+				       "blink")
+		   (list x))
 	    (exact-match "inherit" x)))
      ("none") #f)
     ("text-indent" ,(matchf (string-append length-regex "|" percentage-regex 
@@ -583,6 +596,27 @@
 (define null-merge 
   (lambda (x y) (if (not (null? x)) (if (null? y) (list x) (cons x y)) y)))
 
+(define guile-unicode-support 
+  (let ((v (string-split (version) #\.)))
+    (or (> 6 (string->number (cadr v)))
+	(> 1 (string->number (car v))))))
+
+(define can-parse-unicode?
+  (lambda (num) (or (< num 256) guile-unicode-support)))
+
+(define replace-escapes
+  (lambda (str)
+    (let ((f (lambda (m)
+	       (let ((esc (substring (match:substring m) 1)))
+		 (if (eqv? (string-ref esc 0) #\\)
+		     "\\"
+		     (let* ((ms (match:substring m 3))
+			    (n (and ms (string->number ms 16))))
+		       (if (and n (can-parse-unicode? n))
+			   (make-string 1 (integer->char n)) 
+			   esc)))))))
+      (regexp-substitute/global #f escape-regexc str 'pre f 'post))))
+
 (define merge-selectors
   (lambda (selector-list)
     (let ((l (list (list))))
@@ -592,8 +626,8 @@
 		      (if (not (null? (car l)))
 			  (set-car! (list-cdr-ref l (- (length l) 1))
 				    (append (list-ref l (- (length l) 1))
-					    (list x)))
-			  (set-car! l (list x)))))
+					    (list (replace-escapes x))))
+			  (set-car! l (list (replace-escapes x))))))
 		selector-list)
       l)))
 
@@ -672,17 +706,19 @@
 
 (define validate-property-parse
   (lambda (property value)
-    (let ((pe (hash-ref property-hash-table property)))
+    (let* ((escd-prop (replace-escapes property))
+	   (escd-val (replace-escapes value))
+	   (pe (hash-ref property-hash-table escd-prop)))
       (if pe
-	  (let ((v2 (let ((x (string-length value)))
-		      (if (and (> x 11) (equal? (substring value (- x 10))
+	  (let ((v2 (let ((x (string-length escd-val)))
+		      (if (and (> x 11) (equal? (substring escd-val (- x 10))
 						"!important"))
-			  (substring value 0 (- x 11))
-			  value))))
+			  (substring escd-val 0 (- x 11))
+			  escd-val))))
 	    (if (apply (car pe) (list (string-downcase v2)))
-		(cons property (check-important value))
+		(cons escd-prop (check-important escd-val))
 		(list)))
-	  (list property value)))))
+	  (list escd-prop escd-val)))))
 
 (define validate-property 
   (lambda (property value)
@@ -1006,7 +1042,7 @@
 			  ml (list-index (lambda (y) (equal? y d))
 					 '("top" "right" "bottom" "left"))))
 		    (else #f)))))
-	   (h (lambda (str pattern)
+ 	   (h (lambda (str pattern)
 		(let ((match (if (string? pattern) 
 				 (string-match pattern str)
 				 (regexp-exec pattern str))))
@@ -1294,7 +1330,7 @@
 				(get-more-general-properties prop)))
 		     (lambda (x y) 
 		       (or (not y) (and x (> (cddr x) (cddr y)))))))
-	   (v (if (car vs)
+	   (v (if (car vs)		  
 		  (if (equal? (caar vs) prop)
 		      (cadar vs)
 		      (get-general-property prop 
